#!/usr/bin/env python3
"""
This file is part of cheap_pie, a python tool for chip validation
 author: Marco Merlin
 email: marcomerli@gmail.com
"""

if __name__ == '__main__':
    # needed if cheap_pie not installed
    import os
    import sys
    sys.path.append(os.path.join(os.path.dirname(__file__), '../..'))

from cheap_pie.parsers.cp_parsers_wrapper import cp_parsers_wrapper  # pylint: disable=C0413,E0401
from cheap_pie.parsers.svd_parse_repo import svd_repo_print_vendors, svd_repo_print_vendor_devices # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_banner import cp_banner            # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_cli import cp_cli                  # pylint: disable=C0413,E0401
from cheap_pie.transport.cp_dummy_transport import CpDummyTransport  # pylint: disable=C0413,E0401

# Ipython autoreload
# %load_ext autoreload
# %autoreload 2
# %reset
# %run cheap_pie


def cp_main(argv=[]):  # pylint: disable=W0102
    """
    Main cheap_pie function
    """
    ## banner #######################################################################
    cp_banner()
    #
    ## input parameters ###########################################################
    print('Parsing input arguments...')
    prms = cp_cli(argv)

    ## helper functions ###########################################################

    if prms.vendors:
        return svd_repo_print_vendors()
    if prms.devices:
        return svd_repo_print_vendor_devices(vendor=prms.vendor)

    # transport hif interface %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    print('Initialising Host Interface...')

    # init jlink transport: importing under if case allows not installing all transport libraries
    if prms.transport == 'dummy':
        hif = CpDummyTransport()
    elif prms.transport == 'jlink':  # disable jlink for testing
        from cheap_pie.transport.cp_jlink_transport import CpJlinkTransport  # pylint: disable=C0413,C0415,E0401
        hif = CpJlinkTransport(device=prms.device)
    elif prms.transport == 'ocd':
        from cheap_pie.transport.cp_pyocd_transport import CpPyocdTransport  # pylint: disable=C0413,C0415,E0401
        hif = CpPyocdTransport(device=prms.device)
    elif prms.transport == 'esptool':
        from cheap_pie.transport.cp_esptool_transport import CpEsptoolTransport  # pylint: disable=C0413,C0415,E0401
        hif = CpEsptoolTransport(port=prms.port)
    elif prms.transport == 'verilator':
        from cheap_pie.transport.cp_pyverilator_transport import CpPyverilatorTransport  # pylint: disable=C0413,C0415,E0401
        hif = CpPyverilatorTransport(prms.top_verilog)
    else:
        hif = None
        assert False, f'Invalid transport: {prms.transport}'

    ## init chip ##################################################################
    print('Initialising Hardware Abstraction Layer...')
    lhal = cp_parsers_wrapper(prms, hif=hif)

    ## welcome ####################################################################
    print('Cheap Pie is ready! Type hal.<TAB> to start browsing...')

    return lhal


if __name__ == '__main__':
    hal = cp_main(sys.argv[1:])



import sqlite3
import re
import os
from collections import OrderedDict

svd_path = os.path.dirname(__file__)
# Put database file in the same folder and provide name here
db_path = svd_path + "/cube-finder-db.db"
# Table containing part numbers
pn_table = "cpn"
# Column containing part numbers
pn_column = "cpn"

def regexp(expr, item):
    reg = re.compile(expr)
    return reg.search(item) is not None

files = os.listdir(svd_path)
regexes = OrderedDict()
for file in files:
    if file[-4:] == ".svd":
        name=file[0:-4]
        if (name.find("_") != -1):
            name = name[0:name.find("_")]
        # exclude DISCOVERY since otherwise some part numbers match with regular expression
        regexes[file] = "^(?=^" + name.replace("x","[A-Z0-9]") + "[A-Z0-9]*$)(?=^((?!DISCOVERY).)*$).*$"

regexes = OrderedDict(sorted(regexes.items(), key=lambda item: item[1]))

con = sqlite3.connect(db_path)
con.create_function("REGEXP", 2, regexp)
cursor = con.cursor()
contents = [];
for file, regex in regexes.items():
    res = str(cursor.execute("SELECT GROUP_CONCAT(mcus,', ') FROM (SELECT DISTINCT SUBSTRING(cpn, 1, 11) as mcus FROM (SELECT " + pn_column + " FROM " + pn_table + " WHERE " + pn_column + " REGEXP '" + regex + "'))").fetchone())
    if res == "(None,)":
        contents.append(file)
    else:
        contents.append(res[2:-3] + ", " + file)

with open(svd_path + '/Contents.txt', 'w') as f:
    for line in contents:
        f.write(f"{line}\n")

con.close()


#!/usr/bin/env python3

""" JLink Cheap Pie transport module """

# -*- coding: utf-8 -*-
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

import pylink
from cheap_pie.transport.cp_dummy_transport import CpDummyTransport, test_cp_dummy, \
    hifread_preproc, hifwrite_preproc  # pylint: disable=E0401


class CpJlinkTransport(CpDummyTransport):
    """ A wrapper around jlink transport """
    jl = None

    def __init__(self, device=None):
        if not device is None:
            self.jl = pylink.JLink()  # pylint: disable=C0103
            self.jl.exec_command('ProjectFile = JLinkSettings.ini')
            self.jl.open()
            self.jl.set_speed(10)
            self.jl.set_tif(pylink.enums.JLinkInterfaces.SWD)
            self.jl.connect(device)
            assert self.jl.connected()

    def __del__(self):
        if not self.jl is None:
            self.jl.close()

    def hifread(self, addr="0x40000888"):
        """ read register through JLink """

        addr = hifread_preproc(addr)

        if self.jl is None:
            ret = CpDummyTransport.hifread(self, addr)
        else:
            read = self.jl.memory_read32(addr, 1)
            ret = read[0]

        return ret

    def hifwrite(self, addr="0x40000888", val="0x00000352", verify=True):
        """ write register through JLink """

        waddr, wval, _ = hifwrite_preproc(addr, val)

        if self.jl is None:
            CpDummyTransport.hifwrite(self, waddr, wval)
        else:
            self.jl.memory_write32(waddr, [wval])

        if verify:
            assert int(self.hifread(addr))==int(wval)

        return int(wval)


def test_cp_jlink():
    """ test JLink transport """
    test_cp_dummy(
        CpJlinkTransport(device=None)
    )


if __name__ == '__main__':
    test_cp_jlink()



#!/usr/bin/env python3
"""
Cheap Pie Command Line Interface
"""
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

import os
import sys
import pathlib
import argparse


def cp_root_dir():
    """ Returns the full path of Cheap Pie root directory """
    return pathlib.Path(__file__).parent.parent.absolute()


def cp_devices_dir():
    """ Returns the full path of ad filename in Cheap Pie devices directory """
    return os.path.join(
        cp_root_dir(),
        'devices')


def cp_devices_fname(fname):
    """ Returns the full path of ad filename in Cheap Pie devices directory """
    return os.path.join(
        cp_devices_dir(), fname
    )


def cp_cli(args=[]):  # pylint: disable=W0102
    """
    Cheap Pie Command Line Interface
    """
    parser = argparse.ArgumentParser(description='Cheap Pie Configuration')
    # register format options
    parser.add_argument("-rf", "--regfname", help="register description file name",
                        action='store', type=str, default="QN908XC.svd")
    parser.add_argument("-dd", "--devicedir", help="register description files folder",
                        action='store', type=str,
                        default=cp_devices_dir()
                        )
    parser.add_argument("-fmt", "--format", help="device description format",
                        action='store', type=str, default="cmsis-svd",
                        choices=["svd", "cmsis-svd", "ipxact", "ipyxact", "rdl"])
    parser.add_argument("-ve", "--vendor",
                        help="device vendor. if specified parses svd file from github repository.",
                        action='store', type=str, default=None)
    parser.add_argument("-vendors", "--vendors",
                        help="Available device vendors from cmsis-svd package.",
                        action='store_true')
    parser.add_argument("-devices", "--devices",
                        help="Available devices for specified vendor from cmsis-svd package.",
                        action='store_true')
    # transport options
    parser.add_argument("-d", "--device", help="jlink/ocd device name", action='store',
                        type=str, default="CORTEX-M4")
    parser.add_argument("-t", "--transport", help="transport", action='store', type=str,
                        default="dummy", choices=["jlink", "dummy", "ocd", "esptool", "verilator"])
    parser.add_argument("-p", "--port", help="esptool serial port", action='store', type=str,
                        default="/dev/ttyUSB0")
    parser.add_argument("-topv", "--top_verilog",
                        help="top verilog file (when simulating with verilator)",
                        action='store', type=str, default="./devices/rdl/basic/basic_rf.sv")

    return parser.parse_args(args)


def test_cli(argv):
    """
    Test function for Cheap Pie Command Line Interface
    """
    return cp_cli(argv)


if __name__ == '__main__':
    print(test_cli(sys.argv[1:]))



#!/usr/bin/env python3
""" Cheap Pie parser module for .svd files using SVDParser module """
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

# parser build for CMSIS-SVD xml file format
# https://www.keil.com/pack/doc/CMSIS/SVD/html/index.html
#
# simplified manual initialization
#
# import untangle
# hal = untangle.parse('QN908XC.xml')
# from cbitfield import cp_bitfield
# from cp_register import cp_register
# ADC_ANA_CTRL = cp_register("ADC_ANA_CTRL","0x4000702C",
#                            "ADC core and reference setting regsiter" , hif )
# ADC_BM = cp_bitfield("ADC_BM","0x4000702C",
#                      "ADC_ANA_CTRL",3,0, "ADC bias current selection." ,hif)

import os
from ast import literal_eval
from cmsis_svd.parser import SVDParser

from cheap_pie.cheap_pie_core.cp_builder import CpHalBuilder  # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_cli import cp_devices_fname  # pylint: disable=C0413,E0401

import os
import requests
import zipfile
from urllib.parse import urlparse

def download_github_repo(repo_url, output_dir="."):
    """
    Download a GitHub repository as a ZIP file and extract it
    
    Args:
        repo_url (str): GitHub repo URL (e.g., "https://github.com/user/repo")
        output_dir (str): Directory to save the repository
    """
    try:
        # Parse the repository URL
        parsed = urlparse(repo_url)
        if not parsed.netloc == "github.com":
            raise ValueError("URL must be from github.com")
        
        path_parts = parsed.path.strip("/").split("/")
        if len(path_parts) < 2:
            raise ValueError("Invalid GitHub repository URL")
        
        owner, repo = path_parts[:2]
        
        # Create output directory if it doesn't exist
        os.makedirs(output_dir, exist_ok=True)
        
        # Download the ZIP file
        zip_url = f"https://github.com/{owner}/{repo}/archive/refs/heads/main.zip"
        print(f"Downloading {zip_url}...")
        
        response = requests.get(zip_url, stream=True)
        response.raise_for_status()
        
        zip_path = os.path.join(output_dir, f"{repo}.zip")
        
        with open(zip_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        
        # Extract the ZIP file
        print(f"Extracting to {output_dir}...")
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(output_dir)
        
        # Remove the ZIP file
        os.remove(zip_path)
        
        print(f"Successfully downloaded {owner}/{repo} to {output_dir}")
        return os.path.join(output_dir, f"{repo}-main")
        
    except Exception as e:
        print(f"Error downloading repository: {e}")
        return None

def get_cmsis_svd_data_dir():
    outdir = "."
    datalib = os.path.join(outdir,'cmsis-svd-data-main','data')
    if not os.path.isdir(datalib):
        download_github_repo(repo_url='https://github.com/cmsis-svd/cmsis-svd-data', output_dir=outdir)
    assert os.path.isdir(datalib)
    return datalib

def svd_parse_repo(fname, vendor=None, hif=None, base_address_offset="0x00000000",
                   svd_root=get_cmsis_svd_data_dir()
                   ):
    """ Cheap Pie parser function for .svd files using SVDParser module """
    ## read input file ########################################################
    if vendor is None:
        svd = SVDParser.for_xml_file(fname)
    else:
        svd = SVDParser.for_packaged_svd(package_root=svd_root, vendor=vendor, filename=fname)

    ## loop over lines ########################################################
    cpb = CpHalBuilder(hif)

    for periph in svd.get_device().peripherals:  # pylint: disable=R1702
        # print(periph.name.cdata)

        if hasattr(periph, 'registers'):
            for reg in periph.registers:
                if hasattr(reg, 'name'):
                    # new register
                    regaddr = (
                        reg.address_offset +
                        periph.base_address +
                        literal_eval(base_address_offset)
                    )

                    if hasattr(reg, 'reset_value'):
                        reset=reg.reset_value
                    else:
                        reset=0

                    cpb.reg_open(
                        regname=f'{periph.name}_{reg.name}',
                        regaddr=regaddr,
                        comments=reg.description,
                        reset=reset
                    )

                    if hasattr(reg, 'fields'):
                        for field in reg.fields:
                            if not field is None:
                                # reset value, if available
                                if hasattr(field, 'reset_value'):
                                    reset=field.reset_value
                                else:
                                    reset=0

                                # Create new field class
                                cpb.newfield(
                                    regfield=field.name,
                                    width=field.bit_width,
                                    offset=field.bit_offset,
                                    comments=field.description,
                                    reset=reset
                                )

    # convert output dictionary into structure
    return cpb.out()

def svd_repo_list_vendors() -> list:
    """ List the available vendors for function svd_parse_repo """
    data_path = get_cmsis_svd_data_dir()
    assert os.path.isdir(data_path), f'ERROR: directory {data_path} does not exist!'
    return [ f.path.split('/')[-1] for f in os.scandir(data_path) if f.is_dir() ]

def svd_repo_print_vendors() -> list:
    """ Print the available vendors for function svd_parse_repo """
    vendors = svd_repo_list_vendors()
    print('Available vendors:')
    for v in vendors:
        print(v)
    return vendors

def svd_repo_list_vendor_devices(vendor='Freescale') -> list:
    """ List the available devices for function svd_parse_repo """
    data_path = os.path.join(get_cmsis_svd_data_dir(),vendor)
    assert os.path.isdir(data_path), f'ERROR: directory {data_path} does not exist!'
    return [ f.path.split('/')[-1] for f in os.scandir(data_path) if f.is_file() ]

def svd_repo_print_vendor_devices(vendor='Freescale') -> list:
    """ Print the available devices for function svd_parse_repo """
    devices = svd_repo_list_vendor_devices(vendor=vendor)
    print(f'Available devices for vendor {vendor}:')
    for d in devices:
        print(d)
    return devices

def test_svd_parse_repo():
    """ Test Function for .svd parser based of SVDParser module """
    print('Testing QN9080 with repo parser...')

    fname = cp_devices_fname("QN908XC.svd")
    hal = svd_parse_repo(fname=fname)
    assert len(hal) > 0

    print('Testing K20 with repo parser...')
    hal = svd_parse_repo(fname='MK20D7.svd', vendor='Freescale')
    assert len(hal) > 0

    print('Testing svd_repo_list_vendors...')
    vendors = svd_repo_list_vendors()
    print(vendors)
    assert len(vendors) > 0

    print('Testing svd_repo_print_vendors...')
    vendors = svd_repo_print_vendors()
    print(vendors)
    assert len(vendors) > 0

    print('Testing svd_repo_list_vendor_devices...')
    devices = svd_repo_list_vendor_devices()
    print(devices)
    assert len(devices) > 0

    print('Testing svd_repo_print_vendor_devices...')
    devices = svd_repo_print_vendor_devices()
    print(devices)
    assert len(devices) > 0

if __name__ == '__main__':
    test_svd_parse_repo()



#!/usr/bin/env python3
""" Cheap Pie module to parse .rdl files """
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

from cheap_pie.parsers.ipyxact_parse import ipyxact_parse    # pylint: disable=C0413,E0401
from cheap_pie.tools.rdl2any import rdl2any                  # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_cli import cp_devices_fname  # pylint: disable=C0413,E0401


def rdl_parse(fname, hif=None, base_address_offset="0x00000000"):
    """ Cheap Pie function to parse .rdl files """

    # convert rdl to ipxact
    ipxact_fname = rdl2any(['-f', fname, '-ofmt', 'ipxact'])
    # parse ipxact
    return ipyxact_parse(ipxact_fname, hif=hif, base_address_offset=base_address_offset)


def test_rdl_parse():
    """ Cheap Pie function to test .rdl file parser """
    rdl_parse(cp_devices_fname("rdl/basic.rdl"))
    rdl_parse(cp_devices_fname("rdl/counter.rdl"))


if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        FNAME = sys.argv[1]
    else:
        FNAME = cp_devices_fname("rdl/basic.rdl")
    print(FNAME)
    hal = rdl_parse(FNAME)
    print(hal)



#!/usr/bin/env python3
"""
Register Class Module for Cheap Pie
"""
#
# -*- coding: utf-8 -*-
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

import textwrap
from collections import namedtuple
from ast import literal_eval
from operator import attrgetter
import json

import wavedrom
from wavedrom_ascii import BitfieldASCII
from cheap_pie.cheap_pie_core.cbitfield import CpBitfield


def dict2namedtuple(outdict, tuplename="HAL"):
    """ Convert a dictionary into a namedtuple """
    assert isinstance(outdict, dict)
    return namedtuple(tuplename, outdict.keys())(*outdict.values())


def isnamedtupleinstance(xtuple):
    """ check if a given instance is a namedtuple (can return false positives) """
    # https://stackoverflow.com/questions/2166818/how-to-check-if-an-object-is-an-instance-of-a-namedtuple
    ttt = type(xtuple)
    bbb = ttt.__bases__
    if len(bbb) != 1 or bbb[0] != tuple:
        return False
    fff = getattr(ttt, '_fields', None)
    if not isinstance(fff, tuple):
        return False
    return all(isinstance(n, str) for n in fff)


def reg_add_reserved_bitfields(fields, regwidth=32, fieldname="Reserved", read_write="r"):
    """ Add inferred reserved bits to register description """
    next_lsb = 0
    outfields = []

    if isinstance(fields, list):
        if len(fields) > 0:
            for idx in reversed(range(len(fields))):
                field = fields[idx]
                outfields.insert(0, field)
                # outfields.append(field)

                if next_lsb < field.lsb:
                    # print("Add Reserved field!")
                    newfield = CpBitfield(fieldname, 0, field.regname,
                                          field.lsb-next_lsb, next_lsb,
                                          comments="Reserved",
                                          read_write=read_write)
                    outfields.insert(1, newfield)
                    # outfields.append(newfield)

                # print("LSB: %d, WIDTH: %d" % ( field.lsb, field.width ) )
                next_lsb = field.lsb + field.width

            # check if register is filled up to full width
            if next_lsb < regwidth:
                newfield = CpBitfield(fieldname, 0, fields[0].regname,
                                      regwidth-next_lsb, next_lsb,
                                      comments="Reserved",
                                      read_write=read_write
                                      )
                outfields.insert(0, newfield)

    return outfields


class CpRegister():  # pylint: disable=R0902
    """
    Register Class for Cheap Pie
    """
    addr = 0
    regname = ''
    bitfields = []
    comments = ''
    # host interface handler
    hif = None
    addr_offset = 0
    addr_base = 0
    reset = 0 
    #

    def __init__(self, 
                 regname, 
                 regaddr, 
                 comments, 
                 hif, 
                 addr_offset=0, 
                 addr_base=0, 
                 bitfields=None,
                 reset=0
                 ):  # pylint: disable=R0913
        # address
        self.addr = regaddr

        # name
        self.regname = regname

        # fields
        if not bitfields is None:
            self.bitfields = dict2namedtuple(bitfields)

        # Comments
        self.comments = comments

        # host interface handler
        self.hif = hif

        # address offset
        self.addr_offset = addr_offset

        # address base
        self.addr_base = addr_base

        # reset value
        self.reset = reset

    def getreg(self, asdict=False, as_signed=False, **kwargs):
        """ function getreg(self,regval)
        %
        % Displays value of a register from a register value
        %
        % input : regval value of the full register either in decimal or
        % hexadecimal """

        if not self.hif is None:
            regval = self.hif.hifread(self.addr, **kwargs)
        else:
            regval = 0

        if asdict:
            # dictionary
            retval = {}
            for field in self.bitfields:
                retval[field.fieldname] = field.getbit(regval)
        elif as_signed:
            # signed integer
            retval = -(regval & literal_eval('0x80000000')) + \
                (regval & literal_eval('0x7FFFFFFF'))
        else:
            # unsigned integer (default)
            retval = regval

        return retval

    def setreg(self, regval=0, echo=False, verify=True, **kwargs):  # pylint: disable=W1113
        """ function setreg(self,regval)
        %
        % Displays value of a register from a register value
        %
        % input : regval value of the full register either in decimal or
        % hexadecimal """

        ## handle string input as binary #################################################
        if isinstance(regval, str):
            regval = literal_eval(regval)

        ## handle dict values ############################################################
        elif isinstance(regval, dict):
            rval = self.getreg()
            for field, fval in regval.items():
                rval = self[field].setbit(fval, regval=rval, writeback=False)
            regval = rval

        ## handle negative values ########################################################
        elif regval < 0:
            regval = (abs(regval) ^ literal_eval('0xFFFFFFFF')) + 1
            # print('regval write: 0x%x' % regval)

        # % write %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        if not self.hif is None:
            ret = self.hif.hifwrite(self.addr, regval, verify=verify, **kwargs)
        else:
            ret = regval

        # % only display output if no nargout %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if echo:
            self.display(regval)

        return ret

    def getbit(self, bit_offset=0, width=1, **kwargs):
        """ Get a custom bitfield within a register """
        bitfield = CpBitfield(regaddr=self.addr, width=width,
                              bit_offset=bit_offset, hif=self.hif,
                              **kwargs)
        return bitfield.getbit()

    def setbit(self, bitval=0, bit_offset=0, width=1, verify=True,
               **kwargs):
        """ Set a custom bitfield within a register """
        bitfield = CpBitfield(regaddr=self.addr, width=width,
                              bit_offset=bit_offset, hif=self.hif)
        return bitfield.setbit(bitval, verify=verify, **kwargs)

    def getbyte(self, byte_offset=0, **kwargs):
        """ Get a custom byte within a register """
        byte = CpBitfield(regaddr=self.addr, width=8,
                          bit_offset=byte_offset*8, hif=self.hif)
        return byte.getbit(**kwargs)

    def setbyte(self, byteval=0, byte_offset=0, verify=True,
                **kwargs):
        """ Set a custom byte within a register """
        byte = CpBitfield(regaddr=self.addr, width=8,
                          bit_offset=byte_offset*8, hif=self.hif)
        return byte.setbit(byteval, verify=verify, **kwargs)

    def get_addr(self, mode='hex'):
        """ Returns the address of the register
          input:
            mode: 'hex' return hex string, 'dec' returns decimal number
        """
        assert mode in ['hex','dec'], f"Wrong mode {mode}"
        if mode=='dec':
            return self.addr
        if mode=='hex':
            return hex(self.addr) # return hex string
        return 0

    def get_reset(self, mode='hex'):
        """ Returns the reset value of the register
          input:
            mode: 'hex' return hex string, 'dec' returns decimal number
        """
        assert mode in ['hex','dec'], f"Wrong mode {mode}"
        if mode=='dec':
            return self.reset
        if mode=='hex':
            return hex(self.reset) # return hex string
        return 0

    def help(self, width=25):
        """ function ret = help(self)
        # displays register comments """
        self.print_wavedrom()
        print(f"Address: {self.get_addr()}, Reset: {self.get_reset()}")
        print(self.comments)

        fmtstr = '%%%ds: %%s' % width  # pylint: disable=C0209
        for field in self.bitfields:
            print(fmtstr % (field.fieldname, ''))

            for line in textwrap.wrap(field.comments):
                print(fmtstr % ('', line))

            print(fmtstr % ('', f'Reset: {field.get_reset()}'))

    def __repr__(self, regval=None):
        if len(self.bitfields) > 0:
            reg = []
            for field in self.bitfields:
                # field.display(regval)
                reg.append(field.__repr__(regval=regval))
            outstr = "\n".join(reg)
        else:
            outstr = self.regname + ' = '
            if not regval is None:
                outstr += hex(regval)
        return outstr

    def display(self, regval=None):
        """ Show a register """
        if regval is None:
            # read register value
            regval = self.getreg()
        outstr = self.__repr__(regval)  # pylint: disable=C2801
        print(outstr)
        # return outstr

    def get_bitfields(self, name=None):
        """
        Find a child element by name
        """
        if name:
            return [e for e in self.bitfields if e._name == name]  # pylint: disable=W0212

        return self.bitfields

    def get_ordered_bitfields(self, fieldname="Reserved"):
        """ Return ordered list of bitfields sorted by position and including Reserved fields """
        # convert bitfields namedtuple into list
        bitfields = list(self.bitfields)

        # sort by lsb
        bitfields = sorted(bitfields, key=attrgetter('lsb'))
        bitfields.reverse()

        # add reserved bitfields
        return reg_add_reserved_bitfields(bitfields, fieldname=fieldname)

    def gen_wavedrom(self, vspace=200):
        """ generate wavedrom representation of register """
        wdfields = []

        for field in reversed(self.get_ordered_bitfields(fieldname="")):
            wdfields.append(
                {'name': field.fieldname,
                 'bits': field.width,
                 'attr': field.read_write,
                 'rotate': -90}
            )

        wdconfig = {'vspace': vspace}

        wdreg = {
            'reg': wdfields,
            'config': wdconfig
        }

        return wdreg

    def save_wavedrom_json(self):
        """ Export wavedrom .svg representation of register """
        jfname = f'{self.regname}.json'

        wdlines = json.dumps(self.gen_wavedrom(), indent=2)

        with open(jfname, 'w', encoding='utf-8') as fileh:
            for line in wdlines:
                fileh.write(line)

        return jfname

    def save_wavedrom_svg(self):
        """ Export wavedrom .svg representation of register """
        svgfname = f'{self.regname}.svg'
        jfname = self.save_wavedrom_json()
        wavedrom.render_file(jfname, svgfname)
        return svgfname

    def print_wavedrom(self):
        """ Display wavedrom representation of register in the terminal """
        field = BitfieldASCII.from_dict(self.gen_wavedrom())
        print('')
        print(field)
        return field

    def __contains__(self, key):
        if self.bitfields is None:
            raise TypeError('not indexable')
        return any(item.fieldname == key for item in self.bitfields)

    def __len__(self):
        return len(self.bitfields)

    def __iter__(self):
        return self.bitfields.__iter__()

    def __next__(self):
        return self.bitfields.next()

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.bitfields[idx]
        if isinstance(idx, str):
            bfdict = self.bitfields._asdict()
            return bfdict[idx]
        assert False, 'Unsupported indexing!'

    def __setitem__(self, idx, value):
        if isinstance(idx, int):
            return self.bitfields[idx].setbit(value)
        if isinstance(idx, str):
            bfdict = self.bitfields._asdict()
            return bfdict[idx].setbit(value)
        assert False, 'Unsupported indexing!'

    def __index__(self):
        return int(self.getreg())


class CpRegBuilder():
    """
    Register Class builder for Cheap Pie
    """
    addr = 0
    regname = ''
    dictfields = {}
    comments = ''
    # host interface handler
    hif = None
    addr_offset = 0
    addr_base = 0
    reset = 0

    def __init__(self, regname, regaddr, comments, hif, addr_offset=0, addr_base=0, reset=0):  # pylint: disable=R0913
        # address
        self.addr = regaddr

        # name
        self.regname = regname

        # Comments
        self.comments = comments

        # host interface handler
        self.hif = hif

        # address offset
        self.addr_offset = addr_offset

        # address base
        self.addr_base = addr_base

        # for some reason need to reset this
        self.dictfields = {}

        # reset value
        self.reset = reset

    def addfield(self, regfield, width, offset, comments='', reset=0):
        """ Add a field to a register dictionary of fields """
        self.dictfields[regfield] = CpBitfield(
            regfield=regfield,
            width=width,
            bit_offset=offset,
            comments=comments,
            regaddr=self.addr,
            regname=self.regname,
            reset=reset,
            hif=self.hif)

    def dictfield2struct(self):
        """ Convert the list of bitfields into a namedtuple """
        return CpRegister(regname=self.regname,
                          regaddr=self.addr,
                          comments=self.comments,
                          hif=self.hif,
                          addr_offset=self.addr_offset,
                          addr_base=self.addr_base,
                          bitfields=self.dictfields,
                          reset=self.reset
                          )


def test_cp_register():  # pylint: disable=R0914,R0915
    """ Cheap Pie test for cp_register class """
    import sys  # pylint: disable=C0415
    import os.path  # pylint: disable=C0415
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    from random import randint  # pylint: disable=C0415
    from transport.cp_dummy_transport import CpDummyTransport  # pylint: disable=C0415,E0401

    reg = CpRegister(
        regname='regname',
        regaddr=10,
        comments='comments',
        hif=CpDummyTransport(),
        addr_offset=10,
        addr_base=10
    )

    print('# setreg, getreg')
    val = 15
    reg.setreg(val)
    assert val == reg.getreg()

    print('# display')
    reg.display()

    print('# help')
    reg.help()

    print('# negative assignement')
    negval = -1
    reg.setreg(negval)
    retval = reg.getreg(as_signed=True)
    print(retval)
    assert retval == negval

    print('# test CpRegBuilder with bitfields')
    reg_build = CpRegBuilder(
        regname='regname',
        regaddr=10,
        comments='comments',
        hif=CpDummyTransport(),
        addr_offset=10,
        addr_base=10
    )

    reg_build.addfield(
        regfield='fname1',
        width='2',
        offset='2',
        comments='comment1',
    )

    reg_build.addfield(
        regfield='fname2',
        width='3',
        offset='4',
        comments='comment2',
    )

    reg = reg_build.dictfield2struct()
    assert isnamedtupleinstance(reg.bitfields)
    assert len(reg.get_bitfields()) == 2

    print('# reg display with bitfields')
    reg.display()
    print('# reg help with bitfields')
    reg.help()

    print('# reg basic access')
    val = 52
    reg.setreg(val)
    assert reg.getreg() == val

    print('# reg basic access with custom args')
    val += 10
    reg.setreg(val, custom_arg = True)
    assert reg.getreg(custom_arg = True) == val

    print('# reg item access')
    reg[0]        # pylint: disable=W0104
    reg['fname1']  # pylint: disable=W0104

    print('# numeric item assignement')
    val = 1
    reg[0] = val
    assert reg[0].getbit() == val

    print('# dict item assignement')
    val = 2
    reg['fname2'] = val
    assert reg['fname2'].getbit() == val

    print('# reg bit access')
    for offset in range(32):
        for bitval in range(2):
            reg.setbit(bitval, bit_offset=offset)
            assert reg.getbit(bit_offset=offset) == bitval

    print('# reg byte access')
    for offset in range(4):
        byteval = randint(0, 255)
        reg.setbyte(byteval, byte_offset=offset)
        assert reg.getbyte(byte_offset=offset) == byteval

    print('# reg integer representation')
    print(hex(reg))

    print('# reg dict-based assignement')
    val1 = 1
    val2 = 2
    dreg = {'fname1': val1, 'fname2': val2}
    reg.setreg(dreg)
    assert reg['fname1'].getbit() == val1
    assert reg['fname2'].getbit() == val2

    print('# reg dict-based readback')
    dregb = reg.getreg(asdict=True)
    assert dreg == dregb

    print('# reg get_ordered_bitfields')
    bitfields = reg.get_ordered_bitfields()
    for field in bitfields:
        print(field)
    assert len(bitfields) == 4

    print('# reg gen wavedrom')
    wd_dict = reg.gen_wavedrom()
    assert isinstance(wd_dict, dict)
    print('# reg save wavedrom json')
    wdfile = reg.save_wavedrom_json()
    assert os.path.isfile(wdfile)
    print('# reg save wavedrom svg')
    wdfile = reg.save_wavedrom_svg()
    assert os.path.isfile(wdfile)

    print('# get_addr default')
    assert reg.get_addr()==hex(reg.addr)
    print('# get_addr hex')
    assert reg.get_addr(mode='hex')==hex(reg.addr)
    print('# get_addr dec')
    assert reg.get_addr(mode='dec')==reg.addr

if __name__ == '__main__':
    test_cp_register()
